#
# cvsdb.py - Implementation of a bsddb/gdbm-compatible CVS interface
# Copyright (C) 2000  Tony Garnock-Jones <tonyg@kcbbs.gen.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import os.path
import fcntl
import string
import sys
import __builtin__

really_use_cvs = 1
no_cvs_debug = 1

if no_cvs_debug:
	stderr_redirection = ' 2>/dev/null'
else:
	stderr_redirection = ''

lockfilename_ = '.lockfile'
defaultsuffix_ = '.item'

class CvsDb:
	def __init__(self, work_dir, filesuffix = defaultsuffix_):
		self.work_dir = os.path.abspath(work_dir)
		self.cache = {}
		self.dirty = {}
		self.lockfile = None
		self.filesuffix = filesuffix
		self.excluded_keys = ['CVS']

	def keys(self):
		if self.filesuffix:
			return map(lambda x, self=self: x[:-len(self.filesuffix)], \
					filter(lambda x, self=self: \
							x[-len(self.filesuffix):] == \
								self.filesuffix, \
						os.listdir(self.work_dir)))
		else:
			lst = self.excluded_keys + [lockfilename_]
			return filter(lambda x, lst=lst: not x in lst, os.listdir(self.work_dir))

	def __len__(self):
		return len(keys(self))

	def tryopen_(self, key, mode = 'r', lockstyle = None, filesuffix = None):
		if not filesuffix:
			filesuffix = self.filesuffix

		try:
			f = __builtin__.open(os.path.join(self.work_dir, key + filesuffix), mode)
		except:
			return None

		if lockstyle:
			fcntl.flock(f.fileno(), lockstyle)
		return f

	def lock(self):
		if not self.lockfile:
			self.lockfile = self.tryopen_(lockfilename_, 'w', fcntl.LOCK_EX, '')

		if self.lockfile:
			return 1
		else:
			return 0

	def unlock(self):
		if self.lockfile:
			self.sync()
			os.unlink(os.path.join(self.work_dir, lockfilename_))
			fcntl.flock(self.lockfile.fileno(), fcntl.LOCK_UN)
			self.lockfile.close()
			self.lockfile = None

	def has_key(self, key):
		if self.cache.has_key(key):
			return 1

		f = self.tryopen_(key)
		if f:
			f.close()
			return 1
		else:
			return 0

	def item_log(self, key):
		if key in self.excluded_keys:
			return None

		try:
			f = os.popen('cd ' + self.work_dir + \
				'; cvs log ' + key + self.filesuffix + stderr_redirection, 'r')
		except:
			return None

		lines = map(string.strip, f.readlines())
		f.close()

		result = []

		i = 0
		while i < len(lines):
			if lines[i] == '----------------------------':
				revision = string.split(lines[i+1], ' ')[1]
				fields = map(string.strip, string.split(lines[i+2], ';'))
				fields = map(lambda x: string.split(x, ':', 1), fields)
				fields = map(lambda x: map(string.strip, x), fields)
				fields = filter(lambda x: len(x) == 2, fields)
				fmap = {}
				for [key, value] in fields:
					fmap[key] = value
				result = result + [(revision, fmap)]
				i = i + 2
			i = i + 1

		return result

	def __getitem__(self, key):
		if self.cache.has_key(key):
			return self.cache[key]

		f = self.tryopen_(key, 'r', fcntl.LOCK_SH)
		if not f:
			raise KeyError(key)

		result = f.read()
		self.cache[key] = result

		fcntl.flock(f.fileno(), fcntl.LOCK_UN)
		f.close()

		return result

	def get_version(self, key, version):
		if key in self.excluded_keys:
			return None

		try:
			f = os.popen('cd ' + self.work_dir + \
				'; cvs update -r ' + version + ' -p ' + \
				key + self.filesuffix + stderr_redirection, 'r')
		except:
			return None

		result = f.read()
		f.close()
		return result

	def current_version(self, key):
		if key in self.excluded_keys:
			return ''

		try:
			f = os.popen('cd ' + self.work_dir + \
				'; cvs status ' + key + self.filesuffix + stderr_redirection, 'r')
		except:
			return None

		lines = f.readlines()
		f.close()

		for line in lines:
			if string.find(line, 'Working revision') != -1:
				line = string.strip(string.split(line, ':')[1])
				line = string.split(line)[0]
				return line

		return ''

	def undirty_(self, key):
		if self.dirty.has_key(key):
			ent = self.dirty[key]
			if ent[0] == 'changed':
				fcntl.flock(ent[1].fileno(), fcntl.LOCK_UN)
				ent[1].close()
			del self.dirty[key]

	def __setitem__(self, key, value):
		self.cache[key] = value

		self.undirty_(key)
		f = self.tryopen_(key, 'w', fcntl.LOCK_EX)

		if not f:
			raise 'Could not create CvsDb key "' + key + '"'

		f.write(value)
		f.flush()
		self.dirty[key] = ('changed', f)

	def __delitem__(self, key):
		if not self.has_key(key):
			raise KeyError(key)

		os.unlink(os.path.join(self.work_dir, key + self.filesuffix))
		self.undirty_(key)
		self.dirty[key] = ('deleted',)

	def close(self):
		self.sync()
		self.unlock()

	def __del__(self):
		self.close()

	def sync(self):
		tocommit = []
		toremove = []
		tounlock = []
		for (key, ent) in self.dirty.items():
			if key in self.excluded_keys:
				continue

			if ent[0] == 'changed':
				tocommit = tocommit + [key + self.filesuffix]
				tounlock = tounlock + [ent[1]]
			elif ent[0] == 'deleted':
				toremove = toremove + [key + self.filesuffix]
			else:
				raise 'Unknown dirty-action: ' + ent[0]

		if really_use_cvs:
			cmd = '( cd ' + self.work_dir + ';'
			action = 0

			if len(toremove):
				action = 1
				cmd = cmd + 'cvs remove ' + string.join(toremove, ' ') + ';'

			if len(tocommit):
				action = 1
				cmd = cmd + 'cvs add -kb ' + string.join(tocommit, ' ') + ';'

			if len(tocommit) or len(toremove):
				action = 1
				cmd = cmd + 'cvs commit -m "CvsDb" ' + \
					string.join(tocommit, ' ') + ' ' + \
					string.join(toremove, ' ')

			cmd = cmd + ' ) >/dev/null'

			if no_cvs_debug:
				cmd = cmd + ' 2>&1'

			if action:
				os.system(cmd)

		for f in tounlock:
			fcntl.flock(f.fileno(), fcntl.LOCK_UN)
			f.close()

		self.dirty = {}
		self.cache = {}

	def exclude_cvs(self, key):
		self.excluded_keys = self.excluded_keys + [key]

def open(work_dir, filesuffix = defaultsuffix_):
	return CvsDb(work_dir, filesuffix)
